#!/usr/bin/env python

import traceback
from multiprocessing import Process, Queue, cpu_count
import tarfile
import shutil
import time
from runner import *

class Runnable(Process):
    def __init__(self, runner, folder, execute_run, time_limit, name='', 
                 do_plot=True, output=None):
        Process.__init__(self)
        self.runner = runner
        self.folder = folder
        self.do_run = execute_run
        self.time_limit = time_limit
        self.name = name
        self.do_plot = do_plot
        self.results = output
        self.multi_process = None

    def run(self):
        #out = 'RunParallel_' + self.name + '_pid' + str(os.getpid()) + '.out'
        out = 'RunParallel_' + self.name + '.out'
        sys.stdout = open(out, 'w')
        sys.stderr = sys.stdout
        try:
            self.__run_runner()
        except:
            print 'Exception in execution of Runner '
            e = sys.exc_info()
            print '    ', e[:2]
            print '    ', traceback.print_tb(e[2], file=sys.stdout)
        sys.stderr.flush()
        sys.stdout.flush()

    def __run_runner(self):
        if self.do_run:
            try: shutil.rmtree(self.folder)
            except: pass
            os.mkdir(self.folder)
        os.chdir(self.folder)
        try:
            if self.do_run:
                res = self.runner.run(self.time_limit, self.multi_process)
            else: 
                res = self.runner.load_results()
            for r in res.values():
                r.process()
                if self.do_plot:
                    r.plot_tid_all()
            if self.results:
                self.results.put(res)
        except: raise
        finally: os.chdir('..')

class Configuration:
    def __init__(self, setup, name, folder, proc_factory=None, w_channel=None):
        self.setup = setup
        self.proc_factory = proc_factory
        self.w_channel = w_channel
        self.name = name
        self.folder = folder

class Stat():
    pass

class MultipleRun():
    def __init__(self, numbers_of_nodes, configs, time_limit, 
                 do_simulation=True, verbose=True):
        self.nums_nodes = numbers_of_nodes
        self.do_simulation = do_simulation
        self.runners = []
        self.results = []
        self.queues = []
        self.statistics = {}
        self.results_ready = False
        for s in configs:
            r = RunnerOverNodes(numbers_of_nodes, s.setup, verbose, 
                                s.proc_factory, s.w_channel)
            q = Queue()
            self.queues.append(q)
            self.results.append(None)
            t = Runnable(r, s.folder, do_simulation, time_limit, s.name, 
                         output=q)
            self.runners.append(t)

    def run_all(self):
        self.statistics = {}
        self.results_ready = False
        c = cpu_count()
        print 'Start ALL simulations: setups=%d, n_nodes=%s' \
              % (len(self.runners), str(self.nums_nodes))
        print 'Using %d processes (multi-core)' % c
        t0 = time.time()
        n_r = len(self.runners)
        if c >= n_r:
            p = int(c) / n_r
            r = int(c) % n_r
            for j in range(n_r):
                self.runners[j].multi_process = p
                if j == 0:
                    self.runners[j].multi_process += r
                self.runners[j].start()
            for j in range(n_r):
                self.results[j] = self.queues[j].get()
                self.runners[j].join()
        else:
            for i in range(0, n_r, c):
                h = min(i + c, n_r)
                for j in range(i, h):
                    self.runners[j].start()
                for j in range(i, h):
                    self.results[j] = self.queues[j].get()
                    self.runners[j].join()
        dt = time.time() - t0
        print 'ALL done in %f seconds (%d:%d:%d)'\
              % (dt, int(dt) / 3600 , (int(dt) / 60) % 60, int(dt) % 60)
        self.results_ready = True
        if not self.do_simulation:
            return
        created_dirs = []
        for r in self.runners:
            created_dirs.append(r.folder)
        self.__create_backup(created_dirs)

    def get_results(self):
        if not self.results_ready:
            raise BaseException('MultipleRun is not ready to plot. '\
                                'Have you called run_all()?')
        ret = map(lambda a, b: (a, sorted(b.items(), key=lambda x: x[0])), 
                  self.runners, self.results)
        return ret

    def __create_backup(self, created_dirs):
        tmp = 'wise_sim_backup_' + time.strftime('%Y%m%d_%H%M%S')
        print 'Creating backup'
        sys.stdout.flush()
        os.mkdir(tmp)
        for folder in created_dirs:
            shutil.copytree(folder, tmp + '/' + folder)
        tar_name = tmp + '.tar.bz2' 
        tar = tarfile.open(tar_name, 'w:bz2')
        tar.add(tmp)
        tar.close()
        shutil.rmtree(tmp)
        print 'Backup file "%s" created.' % tar_name
