#!/usr/bin/env python

import os
import shutil
import math
from setup import *

import utils 
from application import Application_FileContainer, Application_Result, \
                        result_registry

################################################################################
#                            Tracker Base Class                                #
################################################################################
class Tracker_Result(Application_Result):
    def __init__(self):
        self.results = None

    def read_result_file(self, fname):
        parse_sel = int
        parsers = [int, float, int] + ([float] * 100)
        self.results = utils.read_file_to_tables(fname, 7, 
                                                 parse_select=parse_sel, 
                                                 parsers=parsers)

################################################################################
#                             Tracker DPF Classes                              #
################################################################################
class TrackerDPF_FileContainer(Application_FileContainer):
    def __init__(self, tag=''):
        Application_FileContainer.__init__(self, tag)
        self.out_dir = 'scenario_dpf' + self.tag 
        self.result_file = 'dpf_results' + self.tag + '.txt'
        self.part_result_file = 'dpf_part_results' + self.tag + '.txt'

    def get_result_file(self):
        return self.out_dir + '/' + self.result_file

    def get_outfiles_list(self):
        ret = Application_FileContainer.get_outfiles_list(self)
        ret.append(self.result_file)
        ret.append(self.part_result_file)
        folder = self.out_dir + '/'
        return map(lambda f: folder + f, ret)

    def move_files(self):
        d = self.out_dir + '/'
        try:
            shutil.move(d + 'Castalia-Trace.txt', d + self.trace_file)
        except: pass
        try:
            find = os.popen('find . -maxdepth 1 -regex "./[0-9]+-[0-9]+.txt"')
            s_f = find.readline()[:-1] # remove the trailing \n
            shutil.move(d + s_f, d + self.stat_file)
        except: pass
        shutil.move(d + 'dpf_results.txt', d + self.result_file)
        shutil.move(d + 'dpf_part_results.txt', d + self.part_result_file)

    def ready_all(self):
        for f in self.get_outfiles_list():
            if not os.path.isfile(f):
                return False
        return True
    
    def ready_result(self):
        return os.path.isfile(self.out_dir + '/' + self.result_file)

    def prepare_filesystem(self):
        try:
            os.mkdir(self.out_dir)
        except:
            rm_list = self.get_outfiles_list()
            for f in rm_list:
                try:
                    os.remove(f)
                except: pass


class TrackerDPF_Stat():
    pass


class TrackerDPF_Result(Tracker_Result):
    def __init__(self, n_nodes, tag='', sizes=(None, None)):
        Tracker_Result.__init__(self)
        self.n_nodes = n_nodes
        self.archive = TrackerDPF_FileContainer(tag)
        self.field_sizes = sizes
        self.stats = {}

    def has_tid(self, tid):
        if not self.results:
            raise BaseException('Result not loaded. Have you run process()?')
        return (tid in self.results)

    def get_out_dir(self):
        return self.archive.out_dir

    def get_ini_file(self):
        return self.archive.ini_file

    def cleanup(self):
        self.archive.prepare_filesystem()
    
    def collect(self):
        self.archive.move_files()

    def process(self):
        if not self.archive.ready_result():
            raise BaseException('File archive is not ready')
        self.read_result_file(self.archive.get_result_file())
        for tid, result in self.results.items():
            self.__extract_info(tid, result)

    def plot_tid_all(self, limit=None):
        for t in self.results.keys():
            self.plot_tid(t, limit)

    def plot_tid(self, tid, limit=None):
        if not self.has_tid(tid):
            raise BaseException('Requested Target ID not in the results')
        data_fname = 'plot_dpf_tid' + str(tid) + '_data'
        script = 'plotter.gpi'
        os.chdir(self.archive.out_dir)
        try:
            self.__create_gnuplot_tid_data(data_fname, tid, limit)
            self.__create_gnuplot_tid_script(script, data_fname, tid)
            os.system('gnuplot ' + script)
        except: raise
        finally: os.chdir('..')

    def __extract_info(self, tid, result):
        self.stats[tid] = []
        for t in result:
            record = TrackerDPF_Stat()
            record.k = t[0]
            record.true_x = t[3]
            record.true_y = t[4]
            record.p_x = t[5]
            record.p_y = t[6]
            record.t_lost = (t[7] != 0)
            record.delay_a = t[8]
            record.delay_b = t[9]
            record.e_x = record.p_x - record.true_x
            record.e_y = record.p_y - record.true_y
            record.error = math.sqrt((record.e_x * record.e_x) + \
                                     (record.e_y * record.e_y))
            self.stats[tid].append(record)

    def __create_gnuplot_tid_data(self, data_fname, tid, limit):
        with open(data_fname, 'w') as f:
            for cnt, t in enumerate(self.results[tid]):
                ex = t[3] - t[5]
                ey = t[4] - t[6]
                e_all = math.sqrt((ex * ex) + (ey * ey))
                print >> f, "%d %f %f %f %f %f %f %f" %\
                            (t[0], t[3], t[4], t[5], t[6], ex, ey, e_all)
                if limit and cnt >= limit:
                    break

    def __create_gnuplot_tid_script(self, script, data_fname, tid):
        s_x, s_y = self.field_sizes
        if not s_x:
            s_x = 160
        if not s_y:
            s_y = 120
        with open(script, 'w') as f:
            print >> f, 'set term postscript eps dashed color '\
                    'blacktext "Helvetica" 12 '
            print >> f, 'set grid'
            plot_cmd_traj = 'plot '
            plot_cmd_x = 'plot '
            plot_cmd_y = 'plot '
            plot_cmd_err = 'plot '
            plot_cmd_err_x = 'plot '
            plot_cmd_err_y = 'plot '
            plot_cmd_traj += '"' + data_fname + '" using 2:3 with linespoints '\
                          'ps 0.25 pt 7 lw 1 title "Ground Truth", '
            plot_cmd_traj += '"' + data_fname + '" using 4:5 with points '\
                          'ps 0.25 pt 7 title "Estimate"'
            plot_cmd_x += '"' + data_fname + '" using 1:2 with lines lw 2 '\
                          'title "Ground Truth X",'
            plot_cmd_y += '"' + data_fname + '" using 1:3 with lines lw 2 '\
                        'title "Ground Truth Y",'
            plot_cmd_x +=   '"' + data_fname + '" using 1:4 with lines lw 2 '\
                        'title "Nodes X"'
            plot_cmd_y += '"' + data_fname + '" using 1:5 with lines lw 2 '\
                          'title "Node Y"'
            plot_cmd_err += '"' + data_fname + '" using 1:8 with linespoints '\
                            'pt 2 lw 1 title "Error"'
            plot_cmd_err_x += '"' + data_fname + '" using 1:6 with lines lw 2 '\
                              'title "Node X"'
            plot_cmd_err_y += '"' + data_fname + '" using 1:7 with lines lw 2 '\
                              'title "Node Y"'
            eps = 'plot_dpf_tid' + str(tid) + '_trajectory.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, 'set xrange [-10:%f]' % (s_x + 10)
            print >> f, 'set yrange [%f:-10]' % (s_y + 10)
            print >> f, 'set xlabel "X-position"'
            print >> f, 'set ylabel "Y-position"'
            print >> f, plot_cmd_traj;
            print >> f, 'set xlabel "Tracking Step"'
            print >> f, 'set autoscale xy'
            print >> f, 'set ylabel "X-position"'
            eps = 'plot_dpf_tid' + str(tid) + '_x.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, plot_cmd_x;
            print >> f, 'set ylabel "Y-position"'
            eps = 'plot_dpf_tid' + str(tid) + '_y.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, plot_cmd_y;
            print >> f, 'set ylabel "Tracking Error"'
            eps = 'plot_dpf_tid' + str(tid) + '_err.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, plot_cmd_err;
            print >> f, 'set ylabel "X position-error"'
            eps = 'plot_dpf_tid' + str(tid) + '_err_x.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, plot_cmd_err_x;
            print >> f, 'set ylabel "Y position-error"'
            eps = 'plot_dpf_tid' + str(tid) + '_err_y.eps'
            print >> f, 'set output "%s"' % eps
            print >> f, plot_cmd_err_y;
            print >> f, 'quit'

result_registry['WiseCameraDPF'] = TrackerDPF_Result
