#!/usr/bin/env python

import sys
import os
import subprocess
import time
from multiprocessing import Pool
from setup import *
from results  import *

def _run_setup_function((n, sn, res)):
    os.chdir(res.get_out_dir())
    try:
        with open(res.get_ini_file(), 'w') as f:
            print >> f, sn
        cmd = 'Castalia -c General -i ' + res.get_ini_file()
        p = subprocess.Popen(cmd, shell=True, stdout=sys.stdout, 
                             stderr=sys.stderr)
        p.wait()
    except: raise
    finally: os.chdir('..')

class RunnerOverNodes:
    def __init__(self, numbers_of_nodes, top_cam_setup, verbose=False,
                 proc_factory=None, w_channel=None, subdir='..'):
        self.nums_nodes = numbers_of_nodes
        self.setup = top_cam_setup
        if top_cam_setup.app.name not in application.result_registry:
            raise BaseException('Cannot run a simulation for which no result ' \
                                'parser is register.')
        self.results = {}
        self.verbose = verbose
        self.proc_factory = proc_factory
        self.w_channel = w_channel
        self.subdir = subdir

    def run(self, time_limit=None, multi_process=None):
        if not multi_process or multi_process <= 1:
            self.__run_single_process(time_limit)
        else:
            self.__run_multi_process(time_limit, multi_process)
        return self.results
        
    def load_results(self):
        self.results = {}
        for n in self.nums_nodes:
            res = self.__create_result(n)
            self.results[n] = res
        return self.results

    def __run_multi_process(self, time_limit, procs):
        self.__print_header()
        cnt = 1
        self.results = {}
        tmp_list = []
        pool = Pool(processes=procs)
        self.__print('*' * 70)
        self.__print('Running Scenarios in parallel (%d processes)' % procs)
        self.__print('*' * 70)
        self.__print('* Preparing')
        for n in self.nums_nodes:
            res = self.__create_result(n)
            res.cleanup()
            sn = self.__create_network(n, time_limit)
            tmp_list.append((n, sn, res))
        self.__print('* Simulating')
        sys.stdout.flush()
        pool.map(_run_setup_function, tmp_list)
        self.__print('* Collecting results')
        for (n, sn, res) in tmp_list:
            res.collect()
            self.results[n] = res
        self.__print('*' * 70)

    def __run_single_process(self, time_limit):
        self.__print_header()
        cnt = 1
        self.results = {}
        for n in self.nums_nodes:
            res = self.__create_result(n)
            self.__print('*' * 70)
            self.__print('Running Scenario %d/%d : #nodes = %d'\
                         % (cnt, len(self.nums_nodes), n))
            self.__print('*' * 70)
            self.__print('* Preparing')
            res.cleanup()
            self.__print('* Simulating scenario with #nodes = %d' % n)
            self.__run_setup(self.__create_network(n, time_limit), res)
            self.__print('* Collecting results')
            res.collect()
            self.__print('*' * 70)
            self.__print()
            self.results[n] = res
            cnt += 1

    def __create_result(self, n):
        s = (self.setup.field_sizes[0], self.setup.field_sizes[1])
        return application.result_registry[self.setup.app.name](n, 
                                                    '_NODES_' + str(n), sizes=s)

    def __create_network(self, n, tl):
        self.setup.setup(n)
        if self.w_channel and self.proc_factory:
            sn = self.setup.create_network(self.proc_factory, 
                                           self.w_channel, tl)
        elif self.proc_factory:
            sn = self.setup.create_network(self.proc_factory, 
                                           time_limit=tl)
        elif self.w_channel:
            sn = self.setup.create_network(w_channel=self.w_channel, 
                                           time_limit=tl)
        else:
            sn = self.setup.create_network(time_limit=tl)
        sn.set_subdir(self.subdir)
        return sn

    def __run_setup(self, sn, res):
        os.chdir(res.get_out_dir())
        try:
            with open(res.get_ini_file(), 'w') as f:
                print >> f, sn
            cmd = 'Castalia -c General -i ' + res.get_ini_file()
            self.__print('* Executing command "%s"' % cmd)
            sys.stdout.flush()
            t0 = time.time()
            p = subprocess.Popen(cmd, shell=True, stdout=sys.stdout, 
                                 stderr=sys.stderr)
            p.wait()
            dt = time.time() - t0
            self.__print('* Simulation done in %f seconds (%d:%d:%d)'\
                     % (dt, int(dt) / 3600 , (int(dt) / 60) % 60, int(dt) % 60))
            sys.stdout.flush()
        except: raise
        finally: os.chdir('..')

    def __print_header(self):
        self.__print()
        self.__print('-' * 80)
        self.__print('    START SIMULATIONS: %d scenarios with '\
                     'variable number of nodes' % len(self.nums_nodes))
        self.__print('-' * 80)
        self.__print()
        self.__print()

    def __print(self, msg=''):
        if not self.verbose:
            return
        print msg
